/****************************************************************************\

   BUGLITE.C : Bug Lite demo
   Source code by Diana Gruber
   Artwork by Andy Hunter
   Copyright 1995-1998 Ted Gruber Software, Inc.

   Requires Win32 and optionally DirectX

\****************************************************************************/

// The DIRECTDRAW symbol must be defined if linking with FGW's DirectDraw
// libraries, and commented out if linking with FGW's native Win32 libraries.

#define DIRECTDRAW


// The WINDOWED symbol must be defined if creating a windowed program, and
// commented out if creating a full screen DirectDraw program.

//#define WINDOWED


// The FULL_SCREEN_BLIT and FULL_SCREEN_FLIP symbols are mutually exclusive
// and apply only to full screen DirectDraw programs. Define FULL_SCREEN_BLIT
// to use DirectDraw blitting, or define FULL_SCREEN_FLIP to use DirectDraw
// page flipping. Comment out both symbols if WINDOWED is defined above.

#ifdef DIRECTDRAW
#define FULL_SCREEN_BLIT
//#define FULL_SCREEN_FLIP
#endif


// The WIDTH and HEIGHT symbols define the virtual buffer resolution in
// pixels. For full screen DirectDraw programs, they also define the screen
// resolution. Normally they should be left at 640x480.

#define WIDTH  640
#define HEIGHT 480

/****************************************************************************/

#include "bugdefs.h"

BOOL AppIsActive = FALSE;

int PASCAL WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR lpszCmdParam, int nCmdShow)
{
   static char szAppName[] = "Bug Lite";
   HWND        hwnd;
   MSG         msg;
   WNDCLASS    wndclass;

   if (!hPrevInstance)
   {
      wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
      wndclass.lpfnWndProc   = WndProc;
      wndclass.cbClsExtra    = 0;
      wndclass.cbWndExtra    = 0;
      wndclass.hInstance     = hInstance;
      wndclass.hIcon         = LoadIcon(NULL,IDI_APPLICATION);
      wndclass.hCursor       = (HBRUSH)LoadCursor(NULL,IDC_ARROW);
      wndclass.hbrBackground = NULL;
      wndclass.lpszMenuName  = NULL;
      wndclass.lpszClassName = szAppName;
      RegisterClass(&wndclass);
   }

#ifdef WINDOWED
   hwnd = CreateWindow(
      szAppName,               // window class name
      "Bug Lite",              // window caption
      WS_OVERLAPPEDWINDOW,     // window style
      0,                       // initial x position
      0,                       // initial y position
      WIDTH,                   // initial x size
      HEIGHT,                  // initial y size
      NULL,                    // parent window handle
      NULL,                    // window menu handle
      hInstance,               // program instance handle
      NULL);                   // creation parameters
#else
   hwnd = CreateWindowEx(
      WS_EX_TOPMOST,           // extended window style
      szAppName,               // window class name
      "Bug Lite",              // window caption
      WS_POPUP,                // window style
      0,                       // initial x position
      0,                       // initial y position
      WIDTH,                   // initial x size
      HEIGHT,                  // initial y size
      NULL,                    // parent window handle
      NULL,                    // window menu handle
      hInstance,               // program instance handle
      NULL);                   // creation parameters
#endif

   ShowWindow(hwnd,nCmdShow);
   UpdateWindow(hwnd);

   /* animate those bugs! */

   while (TRUE)
   {
      if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))
      {
         if (msg.message == WM_QUIT)
            break;
         else
         {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
         }
      }
      else if (AppIsActive)
         animate();
   }

   return msg.wParam;
}

/*********************************************************************/

long WINAPI _export WndProc(HWND hwnd, UINT message,
                            UINT wParam, LONG lParam)
{
   static HDC hdc;

   switch (message)
   {
      case WM_ACTIVATEAPP:
         AppIsActive = wParam;
         return 0;

      case WM_CREATE:
#ifdef FULL_SCREEN_BLIT
         fg_ddsetup(WIDTH,HEIGHT,8,0);
#endif
#ifdef FULL_SCREEN_FLIP
         fg_ddsetup(WIDTH,HEIGHT,8,1);
#endif
         hdc = GetDC(hwnd);
         fg_setdc(hdc);
         hpal = fg_defpal();
         fg_realize(hpal);
         ShowWindow(hwnd,SW_SHOWNORMAL);

         fg_vbinit();
#ifdef DIRECTDRAW
         if (fg_ddstatus() != 0)
         {
            MessageBox(hwnd,"DirectDraw setup error.","Error",MB_OK|MB_ICONSTOP);
            DestroyWindow(hwnd);
            return 0;
         }
#endif

         /* create the workspace buffer (or use DirectDraw's back buffer) */
#ifdef FULL_SCREEN_FLIP
         workvb = 0;
#else
         workvb = fg_vballoc(WIDTH,HEIGHT);
#endif
         fg_vbopen(workvb);
         fg_vbcolors();

         /* create a virtual buffer that holds a clean copy of background */
         storevb = fg_vballoc(WIDTH,HEIGHT);
         fg_vbopen(storevb);
         fg_vbcolors();

         /* display the background */
         fg_showpcx("BUGBACK.PCX",8);

         /* load up everything */
         load_bugs();
         init_globals();

#ifndef WINDOWED
         /* hide mouse cursor for a full screen DirectDraw program */
         fg_mouseini();
         fg_mousevis(0);
#endif
         /* copy background from storage buffer to workspace buffer */
         fg_copypage(storevb,workvb);

#ifdef FULL_SCREEN_FLIP
         /* copy background to both the front buffer and the back buffer */
         fg_ddflip();
         fg_copypage(storevb,workvb);
#endif
         return 0;

      case WM_KEYDOWN:
         switch(wParam)
         {
            case VK_ESCAPE:
            case VK_F12:
               DestroyWindow(hwnd);
               break;
         }
         break;

      case WM_SETFOCUS:
         fg_realize(hpal);
         return 0;

      case WM_SIZE:
         cxClient = LOWORD(lParam);
         cyClient = HIWORD(lParam);
         return 0;

      case WM_DESTROY:
#ifndef WINDOWED
         fg_mousevis(1);
#endif
         fg_vbclose();
         fg_vbfree(storevb);
#ifndef FULL_SCREEN_FLIP
         fg_vbfree(workvb);
#endif
         fg_vbfin();
         DeleteObject(hpal);
         ReleaseDC(hwnd,hdc);
         PostQuitMessage(0);
         return 0;
   }
   return DefWindowProc(hwnd,message,wParam,lParam);
}

/*********************************************************************/
void animate()
{
   OBJp node;
   OBJp next_node;

   if (IRAND(0,40) == 9 && nbugs < 5)
      new_bug();

   /* repair the screen under all the bugs */
   for (node = bottom_node; node != (OBJp)NULL; node = next_node)
   {
      next_node = node->next;
      repair_screen(node);
   }

   /* move the bugs */
   for (node = bottom_node; node != (OBJp)NULL; node = next_node)
   {
      next_node = node->next;
      node->action(node);
   }

   /* display all the bugs */
   for (node = bottom_node; node != (OBJp)NULL; node = next_node)
   {
      next_node = node->next;
      if (node->z < 0)
        put_bug(node);
   }

   put_lamp();

   for (node = bottom_node; node != (OBJp)NULL; node = next_node)
   {
      next_node = node->next;
      if (node->z >= 0)
        put_bug(node);
   }

   put_rail();
   blit();
}

/*********************************************************************/
void blit()
{
#ifdef WINDOWED
   fg_vbscale(0,WIDTH-1,0,HEIGHT-1,0,cxClient-1,0,cyClient-1);
#endif
#ifdef FULL_SCREEN_BLIT
   fg_vbpaste(0,WIDTH-1,0,HEIGHT-1,0,HEIGHT-1);
#endif
#ifdef FULL_SCREEN_FLIP
   fg_ddflip();
#endif
}

/*********************************************************************/
void bug_fall(OBJp objp)
{
   objp->frame++;
   if (objp->frame > 3)
      objp->frame = 0;
   objp->image = mfall[objp->frame];
   objp->y += 14;
   if (objp->y - objp->image->height > HEIGHT)
      objp->action = kill_object;
}

/*********************************************************************/
void bug_go(OBJp objp)
{
   objp->x += objp->xspeed;
   if (objp->x > 700 || objp->x < -60)
      objp->action = kill_object;

   objp->y += objp->yspeed;
   if (objp->y > 360)
      objp->yspeed = -1;
   else if (objp->y < 30)
      objp->yspeed = 1;
   else if (IRAND(0,40) == 9)
     objp->yspeed = IRAND(-2,2);

   objp->frame++;
   if (objp->frame == 7)
   {
      if (IRAND (0,10) == 9)
         objp->action = bug_turn;
   }
   else if (objp->frame > 7)
      objp->frame = 0;

   objp->image = mosquito[objp->frame];

   if (objp->z > -2 && objp->z < 2 && in_zap_range(objp))
   {
      objp->frame = -1;
      objp->action = bug_zap;
   }
}

/*********************************************************************/
void bug_turn(OBJp objp)
{
   objp->frame++;
   if (objp->frame > 7)
      objp->frame = 0;
   else if (objp->frame == 3)
      objp->action = bug_turned;
   objp->image = mturn[objp->frame];
}

/*********************************************************************/
void bug_turned(OBJp objp)
{
   objp->direction = !objp->direction;
   objp->xspeed = -objp->xspeed;
   if (objp->direction == LEFT)
      objp->x -= 36;
   else
      objp->x += 36;

   objp->frame = 0;
   objp->image = mosquito[1];
   if (intersect_lamp(objp) == FALSE)
      objp->z++;
   objp->action = bug_go;
}

/*********************************************************************/
void bug_zap(OBJp objp)
{
   objp->frame++;
   objp->image = mzap[objp->frame];
   if (objp->frame == 3)
   {
      objp->frame = -1;
      objp->action = bug_fall;
   }
}

/*********************************************************************/
void init_globals()
{
   bottom_node = (OBJp)NULL;
   top_node = (OBJp)NULL;
   nbugs = 0;
}

/*********************************************************************/
int intersect_lamp(OBJp objp)
{
   register int xmin, xmax;

   if (objp->y - objp->image->height > 159)
      return(FALSE);
   if (objp->direction == LEFT)
      xmin = objp->x+objp->image->xoffset;
   else
      xmin = objp->x-objp->image->width-objp->image->xoffset;
   xmax = xmin+objp->image->width;

   if (xmax < 260 || xmin > 371)
      return(FALSE);

   return(TRUE);
}

/*********************************************************************/
int in_zap_range(OBJp objp)
{
   register int xmin, xmax;

   if (objp->y - objp->image->height > 120)
      return(FALSE);
   if (objp->direction == LEFT)
      xmin = objp->x+objp->image->xoffset;
   else
      xmin = objp->x-objp->image->width-objp->image->xoffset;
   xmax = xmin+objp->image->width;

   if (xmax < 310 || xmin > 341)
      return(FALSE);

   return(TRUE);
}

/*********************************************************************/
void kill_object(OBJp objp)
{
   OBJp node;

   node = objp;
   if (node == bottom_node)
   {
      bottom_node = node->next;
      if (bottom_node != (OBJp)NULL)
         bottom_node->prev = (OBJp)NULL;
   }
   else if (node == top_node)
   {
      top_node = node->prev;
      top_node->next = (OBJp)NULL;
   }
   else
   {
      node->prev->next = node->next;
      node->next->prev = node->prev;
   }
   free(node);
   nbugs--;
}

/*********************************************************************/
void load_bugs()
{
   static int mosq_x1[] = {  5, 72,153,229,  2, 69,150,225};
   static int mosq_x2[] = { 68,148,215,304, 66,145,212,300};
   static int mosq_y1[] = {  0, 34, 33, 17, 96,132,134,114};
   static int mosq_y2[] = { 77, 74, 83, 80,177,175,183,173};
   static int mosq_xo[] = {  0, -5, -2, -4,  0, -5, -2, -4};
   static int mosq_yo[] = {  0,  2, 11,  5,  5,  5, 11,  0};

   static int mfall_x1[] = {  0, 82,153,235};
   static int mfall_x2[] = { 80,150,229,319};
   static int mfall_y1[] = {  0,  0,  0,  0};
   static int mfall_y2[] = { 92, 96,104,105};
   static int mfall_xo[] = { -5, -0, -6, -9};
   static int mfall_yo[] = {  3,  0,  2,  2};

   static int mturn_x1[] = {  3,110,209,  2};
   static int mturn_x2[] = {102,201,311, 68};
   static int mturn_y1[] = {  3,  6, 12, 73};
   static int mturn_y2[] = { 42, 49, 61,152};
   static int mturn_xo[] = {-18,-26,-41,-26};
   static int mturn_yo[] = {  5,  9, 12, 11};

   static int mzap_x1[] = {  0,117,  2,  0};
   static int mzap_x2[] = {111,224, 97,110};
   static int mzap_y1[] = {  0,  0,107,  1};
   static int mzap_y2[] = { 98,106,196,112};
   static int mzap_xo[] = {  5,  3,  0, 12};
   static int mzap_yo[] = {  0, 11,  9,  5};

   register int i;

   fg_vbopen(workvb);
   fg_showpcx("MOSQ.PCX",9);
   for (i = 0; i < 8; i++)
   {
      if ((mosquito[i] = (SPRITE *)malloc(sizeof(SPRITE))) == (SPRITE *)NULL)
      {
         wsprintf(abort_string,"cannot allocate sprites");
      }
      mosquito[i]->height = mosq_y2[i]-mosq_y1[i]+1;
      mosquito[i]->width = mosq_x2[i]-mosq_x1[i]+1;
      mosquito[i]->yoffset = mosq_yo[i];
      mosquito[i]->xoffset = mosq_xo[i];
      fg_move(mosq_x1[i],mosq_y2[i]);

      if ((mosquito[i]->bitmap = (char *)malloc
         (mosquito[i]->height*mosquito[i]->width)) == (char *)NULL)
      {
         wsprintf(abort_string,"cannot allocate bitmaps");
      }
      fg_getimage(mosquito[i]->bitmap,mosquito[i]->width,mosquito[i]->height);
   }

   fg_showpcx("MFALL.PCX",9);
   for (i = 0; i < 4; i++)
   {
      if ((mfall[i] = (SPRITE *)malloc(sizeof(SPRITE))) == (SPRITE *)NULL)
      {
         wsprintf(abort_string,"cannot allocate sprites");
      }
      mfall[i]->height = mfall_y2[i]-mfall_y1[i]+1;
      mfall[i]->width = mfall_x2[i]-mfall_x1[i]+1;
      mfall[i]->xoffset = mfall_xo[i];
      mfall[i]->yoffset = mfall_yo[i];
      fg_move(mfall_x1[i],mfall_y2[i]);

      if ((mfall[i]->bitmap = (char *)malloc
         (mfall[i]->height*mfall[i]->width)) == (char *)NULL)
      {
         wsprintf(abort_string,"cannot allocate bitmaps");
      }
      fg_getimage(mfall[i]->bitmap,mfall[i]->width,mfall[i]->height);
   }

   fg_showpcx("MTURN.PCX",9);
   for (i = 0; i < 4; i++)
   {
      if ((mturn[i] = (SPRITE *)malloc(sizeof(SPRITE))) == (SPRITE *)NULL)
      {
         wsprintf(abort_string,"cannot allocate sprites");
      }
      mturn[i]->height = mturn_y2[i]-mturn_y1[i]+1;
      mturn[i]->width = mturn_x2[i]-mturn_x1[i]+1;
      mturn[i]->xoffset = mturn_xo[i];
      mturn[i]->yoffset = mturn_yo[i];
      fg_move(mturn_x1[i],mturn_y2[i]);

      if ((mturn[i]->bitmap = (char *)malloc
         (mturn[i]->height*mturn[i]->width)) == (char *)NULL)
      {
         wsprintf(abort_string,"cannot allocate bitmaps");
      }
      fg_getimage(mturn[i]->bitmap,mturn[i]->width,mturn[i]->height);
   }

   fg_showpcx("MZAP.PCX",9);
   for (i = 0; i < 3; i++)
   {
      if ((mzap[i] = (SPRITE *)malloc(sizeof(SPRITE))) == (SPRITE *)NULL)
      {
         wsprintf(abort_string,"cannot allocate sprites");
      }
      mzap[i]->height = mzap_y2[i]-mzap_y1[i]+1;
      mzap[i]->width = mzap_x2[i]-mzap_x1[i]+1;
      mzap[i]->xoffset = mzap_xo[i];
      mzap[i]->yoffset = mzap_yo[i];
      fg_move(mzap_x1[i],mzap_y2[i]);

      if ((mzap[i]->bitmap = (char *)malloc
         (mzap[i]->height*mzap[i]->width)) == (char *)NULL)
      {
         wsprintf(abort_string,"cannot allocate bitmaps");
      }
      fg_getimage(mzap[i]->bitmap,mzap[i]->width,mzap[i]->height);
   }

   fg_showpcx("MZAPA.PCX",9);
   if ((mzap[3] = (SPRITE *)malloc(sizeof(SPRITE))) == (SPRITE *)NULL)
   {
      wsprintf(abort_string,"cannot allocate sprites");
   }
   mzap[3]->height = mzap_y2[3]-mzap_y1[3]+1;
   mzap[3]->width = mzap_x2[3]-mzap_x1[3]+1;
   mzap[3]->xoffset = mzap_xo[3];
   mzap[3]->yoffset = mzap_yo[3];
   fg_move(mzap_x1[3],mzap_y2[3]);

   if ((mzap[3]->bitmap = (char *)malloc
      (mzap[3]->height*mzap[3]->width)) == (char *)NULL)
   {
      wsprintf(abort_string,"cannot allocate bitmaps");
   }
   fg_getimage(mzap[3]->bitmap,mzap[3]->width,mzap[3]->height);

   fg_showpcx("LAMP.PCX",9);
   fg_move(12,159);
   fg_getimage(lamp,111,160);

   fg_showpcx("RAIL.PCX",9);
   fg_move(35,199);
   fg_getimage(rail,237,140);
}

/*********************************************************************/
void new_bug()
{
   OBJp node;
   int direction;

   if ((node = (OBJp)malloc(sizeof(OBJ))) == (OBJp)NULL)
       return;

   node->frame = 0;
   node->action = bug_go;
   node->image = mosquito[0];

   direction = IRAND(LEFT,RIGHT);
   if (direction == RIGHT)
   {
      node->x = -60;
      node->xspeed = 4;
   }
   else
   {
      node->x = 700;
      node->xspeed = -4;
   }
   node->direction = direction;

   node->xmin = 0;
   node->xmax = 0;
   node->ymin = 0;
   node->ymax = 0;

   node->prev_xmin = 0;
   node->prev_xmax = 0;
   node->prev_ymin = 0;
   node->prev_ymax = 0;

   node->y = IRAND(40,360);
   node->yspeed = IRAND(-2,2);

   node->z = IRAND(-3,1);

   /* insert this node at the top of the linked list */
   if (bottom_node == (OBJp)NULL)
   {
      bottom_node = node;
      node->prev = (OBJp)NULL;
   }
   else
   {
      node->prev = top_node;
      node->prev->next = node;
   }
   top_node = node;
   node->next = (OBJp)NULL;

   nbugs++;
}

/*********************************************************************/
void put_bug(OBJp objp)
{
   /* put the bug if the bug is pointing left */

   if (objp->direction == LEFT)
   {
      objp->xmin = objp->x+objp->image->xoffset;
      objp->ymax = objp->y+objp->image->yoffset;
      fg_move(objp->xmin,objp->ymax);
      fg_clpimage(objp->image->bitmap,objp->image->width,objp->image->height);
   }

   /* put the bug if the bug is pointing right */
   else
   {
      objp->xmin = objp->x-objp->image->width-objp->image->xoffset;
      objp->ymax = objp->y+objp->image->yoffset;
      fg_move(objp->xmin,objp->ymax);
      fg_flpimage(objp->image->bitmap,objp->image->width,objp->image->height);
   }
   objp->xmax = objp->xmin+objp->image->width;
   objp->ymin = objp->ymax-objp->image->height;
}

/*********************************************************************/
void put_lamp()
{
  fg_move(260,159);
  fg_drwimage(lamp,111,160);
}

/*********************************************************************/
void put_rail()
{
  fg_move(195,479);
  fg_drwimage(rail,237,140);
}

/*********************************************************************/
void repair_screen(OBJp objp)
{
   int xmin, xmax, ymin, ymax;

#ifdef FULL_SCREEN_FLIP
   /* for page flipping, we must repair the screen from one frame back */
   xmin = MAX(0,MIN(objp->xmin,objp->prev_xmin));
   xmax = MIN(WIDTH-1,MAX(objp->xmax,objp->prev_xmax));
   ymin = MAX(0,MIN(objp->ymin,objp->prev_ymin));
   ymax = MIN(HEIGHT-1,MAX(objp->ymax,objp->prev_ymax));

   objp->prev_xmin = objp->xmin;
   objp->prev_xmax = objp->xmax;
   objp->prev_ymin = objp->ymin;
   objp->prev_ymax = objp->ymax;
#else
   /* not page flipping, just repair the screen from current bug position */
   xmin = MAX(0,objp->xmin);
   xmax = MIN(WIDTH-1,objp->xmax);
   ymin = MAX(0,objp->ymin);
   ymax = MIN(HEIGHT-1,objp->ymax);
#endif

   if (xmin <= xmax && ymin <= ymax)
      fg_vbcopy(xmin,xmax,ymin,ymax,xmin,ymax,storevb,workvb);
}
